/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package co.d3s.mca.project01.client;

import co.d3s.mca.project01.common.shittySerialization.ClientData;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.Closeable;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketException;
import java.nio.charset.StandardCharsets;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author 100574275
 */

public class ClientThread extends Thread implements Closeable {
    private String hostname;
    private int port;
    private volatile boolean running;
    DataInputStream inStream;
    DataOutputStream outStream;
    
    //ew I know.
    private ClientForm form;
    
    private Socket socket;
    public ClientThread(String hostname, int port, ClientForm form) {
        this.hostname = hostname;
        this.port = port;
        this.form = form;
        running = false;
    }
    public void run() {
        try {
            socket = new Socket();
            socket.connect(new InetSocketAddress(hostname, port), 30*1000); //Timeout after 30 seconds of doing nothing.
            ClientKeepAliveThread keepAlive = new ClientKeepAliveThread(socket);
            running = true;
            inStream = new DataInputStream(socket.getInputStream());
            outStream = new DataOutputStream(socket.getOutputStream());
            System.out.println("Thread ready and waiting.");
            while(running) {
                int packet = inStream.read();
                System.out.println("Did we read something?");
                switch(packet) {
                    case ClientPacketEnum.LOGIN_REQUEST:
                        System.out.println("Login requested, tell gui!");
                        form.loginPrompt();
                        break;
                    case ClientPacketEnum.LOGIN_RESPONSE:
                        byte type = inStream.readByte();
                        String reason = ReadString();
                        form.loginResponse(type,reason);
                        break;
                    case ClientPacketEnum.KEEP_ALIVE:
                        //We are alive!
                        System.out.println("Received the keep-alive, we are alive!");
                        break;
                    case ClientPacketEnum.FILE_TRANSFER:
                        String fileName = ReadString();
                        int fileLen = inStream.readInt();
                        byte[] fileData = ReadByteArray(fileLen);
                        form.FileTransfer(fileName, fileData);
                        break;
                    case ClientPacketEnum.FILE_LIST:
                        short count = inStream.readShort();
                        String[] files = new String[count];
                        for (int i = 0; i < count; i++) {
                            files[i] = ReadString();
                        }
                        form.fileList(files);
                        break;
                    case ClientPacketEnum.SHITTY_SERIALIZATION:
                        System.out.println("Received a shitty serialization, be careful!");
                        short len = inStream.readShort();
                        byte[] data = ReadByteArray(len);
                        System.out.println("Read everything from the wire");
                        ObjectInputStream test = new ObjectInputStream(new ByteArrayInputStream(data));
                        ClientData object = (ClientData) test.readObject();
                        System.out.println("Read the object and tried to cast it.");
                        form.DataObject(object);
                        break;
                    case ClientPacketEnum.DEBUG:
                        System.out.println("Hey we received a debug packet");
                        String string = ReadString();
                        System.out.println("[DEBUG] " + string);
                        break;
                }
            }
        } catch (SocketException ex) {
            if (ex.getMessage().equals("Socket closed")) {
                System.out.println("We turned off the client, killing the thread.");
                running = false;
            }
        } catch (IOException ex) {
            Logger.getLogger(ClientThread.class.getName()).log(Level.SEVERE, null, ex);
            ex.printStackTrace();
        } catch (Exception ex) {
            Logger.getLogger(ClientThread.class.getName()).log(Level.SEVERE, null, ex);
        }
    }
    public void SendDebugMessage(String string) {
        try {
            outStream.write(ClientPacketEnum.DEBUG);
            WriteString(string);
            outStream.flush();
        } catch (IOException ex) {
            Logger.getLogger(ClientThread.class.getName()).log(Level.SEVERE, null, ex);
        }
    }
    public void RequestFileList() {
        try {
            outStream.write(ClientPacketEnum.FILE_LIST);
            outStream.flush();
        } catch (IOException ex) {
            Logger.getLogger(ClientThread.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    @Override
    public void close() throws IOException {
        socket.close();
        running = false;
    }

    //TODO: Make it encrypted??
    void SendLoginRequest(String username, String password) {
        try {
            outStream.write(ClientPacketEnum.LOGIN_REQUEST);
            WriteString(username);
            WriteString(password);
            outStream.flush();
        } catch (IOException ex) {
            Logger.getLogger(ClientThread.class.getName()).log(Level.SEVERE, null, ex);
        }
    }
    void RequestFile(String file) throws IOException {
        outStream.write(ClientPacketEnum.FILE_REQUEST);
        WriteString(file);
        outStream.flush();
    }
    
    String ReadString() throws IOException, Exception {
        short length = inStream.readShort();
        if (length > -1) {
            int currentLength = 0;
            byte[] pendingString = new byte[length];
            do {
            currentLength += inStream.read(pendingString, currentLength, length - currentLength);
            } while (currentLength < length);
            String string = new String(pendingString, StandardCharsets.UTF_8);
            return string;
        } else {
            throw new Exception("WTF");
        }
    }
    byte[] ReadByteArray(int len) throws IOException {
        int currentLength = 0;
        //Make sure we read this entire byte array.
        byte[] data = new byte[len];
        do {
        currentLength += inStream.read(data, currentLength, len - currentLength);
        } while (currentLength < len);
        return data;
    }
    void WriteString(String string) throws IOException {
        byte[] stringData = string.getBytes(StandardCharsets.UTF_8);
        outStream.writeShort(stringData.length);
        outStream.write(stringData, 0, stringData.length);
    }

    void RequestData() throws IOException {
        outStream.write(ClientPacketEnum.DATA_REQUEST);
        outStream.flush();
    }

    void DeleteData(int intValue) throws IOException {
        outStream.write(ClientPacketEnum.DATA_DELETE);
        outStream.writeInt(intValue);
        outStream.flush();
    }

    void UpdateData(ClientData row) throws IOException {
        outStream.write(ClientPacketEnum.DATA_UPDATE);
        ByteArrayOutputStream outByte = new ByteArrayOutputStream();
        ObjectOutputStream outObject = new ObjectOutputStream(outByte);
        outObject.writeObject(row);
        byte[] bytes = outByte.toByteArray();
        outStream.writeShort(bytes.length);
        outStream.write(bytes);
        outStream.flush();
    }
}
